An input/output controller for virtual pinball machines, with plunger position tracking, accelerometer-based nudge sensing, button input encoding, and feedback device control.

Dependencies:   USBDevice mbed FastAnalogIn FastIO FastPWM SimpleDMA

/media/uploads/mjr/pinscape_no_background_small_L7Miwr6.jpg

The Pinscape Controller is a special-purpose software project that I wrote for my virtual pinball machine.

New version: V2 is now available! The information below is for version 1, which will continue to be available for people who prefer the original setup.

What exactly is a virtual pinball machine? It's basically a video-game pinball emulator built to look like a real pinball machine. (The picture at right is the one I built.) You start with a standard pinball cabinet, either built from scratch or salvaged from a real machine. Inside, you install a PC motherboard to run the software, and install TVs in place of the playfield and backglass. Several Windows pinball programs can take advantage of this setup, including the open-source project Visual Pinball, which has hundreds of tables available. Building one of these makes a great DIY project, and it's a good way to add to your skills at woodworking, computers, and electronics. Check out the Cabinet Builders' Forum on vpforums.org for lots of examples and advice.

This controller project is a key piece in my setup that helps integrate the video game into the pinball cabinet. It handles several input/output tasks that are unique to virtual pinball machines. First, it lets you connect a mechanical plunger to the software, so you can launch the ball like on a real machine. Second, it sends "nudge" data to the software, based on readings from an accelerometer. This lets you interact with the game physically, which makes the playing experience more realistic and immersive. Third, the software can handle button input (for wiring flipper buttons and other cabinet buttons), and fourth, it can control output devices (for tactile feedback, button lights, flashers, and other special effects).

Documentation

The Hardware Build Guide (PDF) has detailed instructions on how to set up a Pinscape Controller for your own virtual pinball cabinet.

Update notes

December 2015 version: This version fully supports the new Expansion Board project, but it'll also run without it. The default configuration settings haven't changed, so existing setups should continue to work as before.

August 2015 version: Be sure to get the latest version of the Config Tool for windows if you're upgrading from an older version of the firmware. This update adds support for TSL1412R sensors (a version of the 1410 sensor with a slightly larger pixel array), and a config option to set the mounting orientation of the board in the firmware rather than in VP (for better support for FP and other pinball programs that don't have VP's flexibility for setting the rotation).

Feb/March 2015 software versions: If you have a CCD plunger that you've been using with the older versions, and the plunger stops working (or doesn't work as well) after you update to the latest version, you might need to increase the brightness of your light source slightly. Check the CCD exposure with the Windows config tool to see if it looks too dark. The new software reads the CCD much more quickly than the old versions did. This makes the "shutter speed" faster, which might require a little more light to get the same readings. The CCD is actually really tolerant of varying light levels, so you probably won't have to change anything for the update - I didn't. But if you do have any trouble, have a look at the exposure meter and try a slightly brighter light source if the exposure looks too dark.

Downloads

  • Config tool for Windows (.exe and C# source): this is a Windows program that lets you view the raw pixel data from the CCD sensor, trigger plunger calibration mode, and configure some of the software options on the controller.
  • Custom VP builds: I created modified versions of Visual Pinball 9.9 and Physmod5 that you might want to use in combination with this controller. The modified versions have special handling for plunger calibration specific to the Pinscape Controller, as well as some enhancements to the nudge physics. If you're not using the plunger, you might still want it for the nudge improvements. The modified version also works with any other input controller, so you can get the enhanced nudging effects even if you're using a different plunger/nudge kit. The big change in the modified versions is a "filter" for accelerometer input that's designed to make the response to cabinet nudges more realistic. It also makes the response more subdued than in the standard VP, so it's not to everyone's taste. The downloads include both the updated executables and the source code changes, in case you want to merge the changes into your own custom version(s).

    Note! These features are now standard in the official VP 9.9.1 and VP 10 releases, so you don't need my custom builds if you're using 9.9.1 or 10 or later. I don't think there's any reason to use my 9.9 instead of the official 9.9.1, but I'm leaving it here just in case. In the official VP releases, look for the checkbox "Enable Nudge Filter" in the Keys preferences dialog. (There's no checkbox in my custom builds, though; the filter is simply always on in those.)
  • Output circuit shopping list: This is a saved shopping cart at mouser.com with the parts needed for each output driver, if you want to use the LedWiz emulator feature. Note that quantities in the cart are for one output channel, so multiply everything by the number of channels you plan to use, except that you only need one of the ULN2803 transistor array chips for each eight output circuits.
  • Lemming77's potentiometer mounting bracket and shooter rod connecter: Sketchup designs for 3D-printable parts for mounting a slide potentiometer as the plunger sensor. These were designed for a particular slide potentiometer that used to be available from an Aliexpress.com seller but is no longer listed. You can probably use this design as a starting point for other similar devices; just check the dimensions before committing the design to plastic.

Features

  • Plunger position sensing, using a TAOS TSL 1410R CCD linear array sensor. This sensor is a 1280 x 1 pixel array at 400 dpi, which makes it about 3" long - almost exactly the travel distance of a standard pinball plunger. The idea is that you install the sensor just above (within a few mm of) the shooter rod on the inside of the cabinet, with the CCD window facing down, aligned with and centered on the long axis of the shooter rod, and positioned so that the rest position of the tip is about 1/2" from one end of the window. As you pull back the plunger, the tip will travel down the length of the window, and the maximum retraction point will put the tip just about at the far end of the window. Put a light source below, facing the sensor - I'm using two typical 20 mA blue LEDs about 8" away (near the floor of the cabinet) with good results. The principle of operation is that the shooter rod casts a shadow on the CCD, so pixels behind the rod will register lower brightness than pixels that aren't in the shadow. We scan down the length of the sensor for the edge between darker and brighter, and this tells us how far back the rod has been pulled. We can read the CCD at about 25-30 ms intervals, so we can get rapid updates. We pass the readings reports to VP via our USB joystick reports.

    The hardware build guide includes schematics showing how to wire the CCD to the KL25Z. It's pretty straightforward - five wires between the two devices, no external components needed. Two GPIO ports are used as outputs to send signals to the device and one is used as an ADC in to read the pixel brightness inputs. The config tool has a feature that lets you display the raw pixel readings across the array, so you can test that the CCD is working and adjust the light source to get the right exposure level.

    Alternatively, you can use a slide potentiometer as the plunger sensor. This is a cheaper and somewhat simpler option that seems to work quite nicely, as you can see in Lemming77's video of this setup in action. This option is also explained more fully in the build guide.
  • Nudge sensing via the KL25Z's on-board accelerometer. Mounting the board in your cabinet makes it feel the same accelerations the cabinet experiences when you nudge it. Visual Pinball already knows how to interpret accelerometer input as nudging, so we simply feed the acceleration readings to VP via the joystick interface.
  • Cabinet button wiring. Up to 24 pushbuttons and switches can be wired to the controller for input controls (for example, flipper buttons, the Start button, the tilt bob, coin slot switches, and service door buttons). These appear to Windows as joystick buttons. VP can map joystick buttons to pinball inputs via its keyboard preferences dialog. (You can raise the 24-button limit by editing the source code, but since all of the GPIO pins are allocated, you'll have to reassign pins currently used for other functions.)
  • LedWiz emulation (limited). In addition to emulating a joystick, the device emulates the LedWiz USB interface, so controllers on the PC side such as DirectOutput Framework can recognize it and send it commands to control lights, solenoids, and other feedback devices. 22 GPIO ports are assigned by default as feedback device outputs. This feature has some limitations. The big one is that the KL25Z hardware only has 10 PWM channels, which isn't enough for a fully decked-out cabinet. You also need to build some external power driver circuitry to use this feature, because of the paltry 4mA output capacity of the KL25Z GPIO ports. The build guide includes instructions for a simple and robust output circuit, including part numbers for the exact components you need. It's not hard if you know your way around a soldering iron, but just be aware that it'll take a little work.

Warning: This is not replacement software for the VirtuaPin plunger kit. If you bought the VirtuaPin kit, please don't try to install this software. The VP kit happens to use the same microcontroller board, but the rest of its hardware is incompatible. The VP kit uses a different type of sensor for its plunger and has completely different button wiring, so the Pinscape software won't work properly with it.

Revision:
26:cb71c4af2912
Parent:
25:e22b88bd783a
Child:
29:582472d0bc57
--- a/main.cpp	Tue Sep 01 04:27:15 2015 +0000
+++ b/main.cpp	Wed Sep 23 05:06:39 2015 +0000
@@ -137,6 +137,15 @@
 //    but with a slight practical need for a handful of extra ports (I'm using the
 //    cutting-edge 10-contactor setup, so my real LedWiz is full!).
 //
+//  - Enhanced LedWiz emulation with TLC5940 PWM controller chips.  You can attach
+//    external PWM controller chips for controlling device outputs, instead of using
+//    the limited LedWiz emulation through the on-board GPIO ports as described above. 
+//    The software can control a set of daisy-chained TLC5940 chips, which provide
+//    16 PWM outputs per chip.  Two of these chips give you the full complement
+//    of 32 output ports of an actual LedWiz, and four give you 64 ports, which
+//    should be plenty for nearly any virtual pinball project.
+//
+//
 // The on-board LED on the KL25Z flashes to indicate the current device status:
 //
 //    two short red flashes = the device is powered but hasn't successfully
@@ -214,6 +223,7 @@
 #include "tsl1410r.h"
 #include "FreescaleIAP.h"
 #include "crc32.h"
+#include "TLC5940.h"
 
 // our local configuration file
 #define DECL_EXTERNS
@@ -226,6 +236,12 @@
 // number of elements in an array
 #define countof(x) (sizeof(x)/sizeof((x)[0]))
 
+// floating point square of a number
+inline float square(float x) { return x*x; }
+
+// floating point rounding
+inline float round(float x) { return x > 0 ? floor(x + 0.5) : ceil(x - 0.5); }
+
 
 // ---------------------------------------------------------------------------
 // USB device vendor ID, product ID, and version.  
@@ -309,6 +325,13 @@
 //
 // On-board RGB LED elements - we use these for diagnostic displays.
 //
+// Note that LED3 (the blue segment) is hard-wired on the KL25Z to PTD1,
+// so PTD1 shouldn't be used for any other purpose (e.g., as a keyboard
+// input or a device output).  (This is kind of unfortunate in that it's 
+// one of only two ports exposed on the jumper pins that can be muxed to 
+// SPI0 SCLK.  This effectively limits us to PTC5 if we want to use the 
+// SPI capability.)
+//
 DigitalOut ledR(LED1), ledG(LED2), ledB(LED3);
 
 
@@ -316,20 +339,77 @@
 //
 // LedWiz emulation
 //
+// There are two modes for this feature.  The default mode uses the on-board
+// GPIO ports to implement device outputs - each LedWiz software port is
+// connected to a physical GPIO pin on the KL25Z.  The KL25Z only has 10
+// PWM channels, so in this mode only 10 LedWiz ports will be dimmable; the
+// rest are strictly on/off.  The KL25Z also has a limited number of GPIO
+// ports overall - not enough for the full complement of 32 LedWiz ports
+// and 24 VP joystick inputs, so it's necessary to trade one against the
+// other if both features are to be used.
+//
+// The alternative, enhanced mode uses external TLC5940 PWM controller
+// chips to control device outputs.  In this mode, each LedWiz software
+// port is mapped to an output on one of the external TLC5940 chips.
+// Two 5940s is enough for the full set of 32 LedWiz ports, and we can
+// support even more chips for even more outputs (although doing so requires
+// breaking LedWiz compatibility, since the LedWiz USB protocol is hardwired
+// for 32 outputs).  Every port in this mode has full PWM support.
+//
 
+// Current starting output index for "PBA" messages from the PC (using
+// the LedWiz USB protocol).  Each PBA message implicitly uses the
+// current index as the starting point for the ports referenced in
+// the message, and increases it (by 8) for the next call.
 static int pbaIdx = 0;
 
 // LedWiz output pin interface.  We create a cover class to virtualize
 // digital vs PWM outputs and give them a common interface.  The KL25Z
 // unfortunately doesn't have enough hardware PWM channels to support 
 // PWM on all 32 LedWiz outputs, so we provide as many PWM channels as
 // we can (10), and fill out the rest of the outputs with plain digital
 // outs.
 class LwOut
 {
 public:
+    // Set the output intensity.  'val' is 0.0 for fully off, 1.0 for
+    // fully on, and fractional values for intermediate intensities.
     virtual void set(float val) = 0;
 };
+
+
+#ifdef ENABLE_TLC5940
+
+// The TLC5940 interface object.
+TLC5940 tlc5940(TLC5940_SCLK, TLC5940_SIN, TLC5940_GSCLK, TLC5940_BLANK,
+    TLC5940_XLAT, TLC5940_NCHIPS);
+
+// LwOut class for TLC5940 outputs.  These are fully PWM capable.
+// The 'idx' value in the constructor is the output index in the
+// daisy-chained TLC5940 array.  0 is output #0 on the first chip,
+// 1 is #1 on the first chip, 15 is #15 on the first chip, 16 is
+// #0 on the second chip, 32 is #0 on the third chip, etc.
+class Lw5940Out: public LwOut
+{
+public:
+    Lw5940Out(int idx) : idx(idx) { prv = -1; }
+    virtual void set(float val)
+    {
+        if (val != prv)
+            tlc5940.set(idx, (int)(val * 4095));
+    }
+    int idx;
+    float prv;
+};
+
+#else // ENABLE_TLC5940
+
+// 
+// Default LedWiz mode - using on-board GPIO ports.  In this mode, we
+// assign a KL25Z GPIO port to each LedWiz output.  We have to use a
+// mix of PWM-capable and Digital-Only ports in this configuration, 
+// since the KL25Z hardware only has 10 PWM channels, which isn't
+// enough to fill out the full complement of 32 LedWiz outputs.
+//
+
+// LwOut class for a PWM-capable GPIO port
 class LwPwmOut: public LwOut
 {
 public:
@@ -342,6 +422,8 @@
     PwmOut p;
     float prv;
 };
+
+// LwOut class for a Digital-Only (Non-PWM) GPIO port
 class LwDigOut: public LwOut
 {
 public:
@@ -354,6 +436,17 @@
     DigitalOut p;
     float prv;
 };
+
+#endif // ENABLE_TLC5940
+
+// LwOut class for unmapped ports.  The LedWiz protocol is hardwired
+// for 32 ports, but we might not want to assign all 32 software ports
+// to physical output pins - the KL25Z has a limited number of GPIO
+// ports, so we might not have enough available GPIOs to fill out the
+// full LedWiz complement after assigning GPIOs for other functions.
+// This class is used to populate the LedWiz mapping array for ports
+// that aren't connected to physical outputs; it simply ignores value 
+// changes.
 class LwUnusedOut: public LwOut
 {
 public:
@@ -361,7 +454,11 @@
     virtual void set(float val) { }
 };
 
-// output pin array
+// Array of output assignments.  This array is indexed by the LedWiz
+// output port number; that protocol is hardwired for 32 ports, so we
+// need 32 elements in the array.  Each element is an LwOut object
+// that provides the mapping to the physical output corresponding to
+// the software port.
 static LwOut *lwPin[32];
 
 // initialize the output pin array
@@ -369,6 +466,18 @@
 {
     for (int i = 0 ; i < countof(lwPin) ; ++i)
     {
+#ifdef ENABLE_TLC5940
+        // Set up a TLC5940 output.  If the output is within range of
+        // the connected number of chips (16 outputs per chip), assign it
+        // to the current index, otherwise leave it unattached.
+        if (i < TLC5940_NCHIPS*16)
+            lwPin[i] = new Lw5940Out(i);
+        else
+            lwPin[i] = new LwUnusedOut();
+
+#else // ENABLE_TLC5940
+        // Set up the GPIO pin, according to whether it's PWM-capable or
+        // digital-only, and whether or not it's assigned at all.
         PinName p = (i < countof(ledWizPortMap) ? ledWizPortMap[i].pin : NC);
         if (p == NC)
             lwPin[i] = new LwUnusedOut();
@@ -376,6 +485,9 @@
             lwPin[i] = new LwPwmOut(p);
         else
             lwPin[i] = new LwDigOut(p);
+            
+#endif // ENABLE_TLC5940
+
     }
 }
 
@@ -601,14 +713,6 @@
 };
 
 // ---------------------------------------------------------------------------
-//
-// Some simple math service routines
-//
-
-inline float square(float x) { return x*x; }
-inline float round(float x) { return x > 0 ? floor(x + 0.5) : ceil(x - 0.5); }
-
-// ---------------------------------------------------------------------------
 // 
 // Accelerometer (MMA8451Q)
 //
@@ -316,20 +339,77 @@
 static int pbaIdx = 0;
 
-// LedWiz output pin interface.  We create a cover class to virtualize
-// digital vs PWM outputs and give them a common interface.  The KL25Z
-// unfortunately doesn't have enough hardware PWM channels to support 
-// PWM on all 32 LedWiz outputs, so we provide as many PWM channels as
-// we can (10), and fill out the rest of the outputs with plain digital
-// outs.
+// Generic LedWiz output port interface.  We create a cover class to 
+// virtualize digital vs PWM outputs, and on-board KL25Z GPIO vs external 
+// TLC5940 outputs, and give them all a common interface.  
 class LwOut
 {